from typing import Union, Tuple
from math import cos, sin
from .utils import Coordinate, Matrix, deg2rad

__all__ = [
    "shearing_matrix",
    "rotation_matrix",
    "scaling_matrix",
    "translation_matrix",
]


def shearing_matrix(angle: float, clockwise: bool = False) -> Matrix:
    r"""Creates an affine horizontal shearing matrix in the form

    .. math::

        \mathrm{\mathbf{S}} =
        \begin{pmatrix}
            1 & - \sin \varphi & 0 \\
            0 & \cos \varphi   & 0 \\
            0 & 0              & 1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}

    Args:
        angle: Angle :math:`\varphi` in degrees.
        clockwise: If ``True``, the shearing will be performed clockwise.
            Defaults to ``False``.

    Returns:
        Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
        :math:`f`.
    """
    angle = deg2rad(angle)
    if clockwise:
        angle *= -1.0
    return (1.0, -sin(angle), 0.0, 0.0, cos(angle), 0.0)


def rotation_matrix(angle: float, clockwise: bool = False) -> Matrix:
    r"""Creates an affine rotation matrix in the form

    .. math::

        \mathrm{\mathbf{R}} =
        \begin{pmatrix}
            \cos \varphi & - \sin \varphi & 0 \\
            \sin \varphi & \cos \varphi   & 0 \\
            0            & 0              & 1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}

    Args:
        angle: Angle :math:`\varphi` in degrees.
        clockwise: If ``True``, the rotation will be performed clockwise.
            Defaults to ``False``.

    Returns:
        Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
        :math:`f`.
    """
    angle = deg2rad(angle)
    if clockwise:
        angle *= -1.0
    return (cos(angle), -sin(angle), 0.0, sin(angle), cos(angle), 0.0)


def scaling_matrix(factor: Union[float, Tuple[float, float]]) -> Matrix:
    r"""Creates an affine scaling matrix in the form

    .. math::

        \mathrm{\mathbf{C}} =
        \begin{pmatrix}
            c_\text{horz} &               & 0 \\
            0             & c_\text{vert} & 0 \\
            0             & 0             & 1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}

    Args:
        factor: Horizontal and vertical scaling factors
            (:math:`c_\text{horz}`, :math:`c_\text{vert}`). If scalar, the
            same factor is used for both directions.

    Returns:
        Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
        :math:`f`.
    """
    if isinstance(factor, float):
        factor_horz = factor_vert = factor
    else:
        factor_horz, factor_vert = factor
    return (factor_horz, 0.0, 0.0, 0.0, factor_vert, 0.0)


def translation_matrix(translation: Coordinate, inverse: bool = False) -> Matrix:
    r"""Creates an affine scaling matrix in the form

    .. math::

        \mathrm{\mathbf{T}} =
        \begin{pmatrix}
            1 & 0 & t_\text{horz} \\
            0 & 1 & t_\text{vert} \\
            0 & 0 & 1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}

    Args:
        translation: Horizontal and vertical translation.
            (:math:`t_\text{horz}`, :math:`t_\text{vert}`)
        inverse: If ``True``, the translation will be performed in the opposite
            direction. Defaults to ``False``.

    Returns:
        Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
        :math:`f`.
    """
    horz_translation, vert_translation = translation
    if inverse:
        horz_translation *= -1.0
        vert_translation *= -1.0
    return (1.0, 0.0, horz_translation, 0.0, 1.0, vert_translation)



from typing import Tuple
from functools import reduce
from math import pi

__all__ = [
    "Coordinate",
    "Matrix",
    "matmul",
    "left_matmuls",
    "matinv",
    "deg2rad",
    "transform_coordinate",
]

Coordinate = Tuple[float, float]
Matrix = Tuple[float, float, float, float, float, float]


def matmul(matrix1: Matrix, matrix2: Matrix) -> Matrix:
    r"""Matrix product

    .. math::

        \begin{pmatrix}
            a_1 & b_1 & c_1 \\
            d_1 & e_1 & f_1 \\
            0   & 0   & 1 \\
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            a_2 & b_2 & c_2 \\
            d_2 & e_2 & f_2 \\
            0   & 0   & 1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}


    Args:
        matrix1: Parameters :math:`a_1`, :math:`b_1`, :math:`c_1`, :math:`d_1`,
            :math:`e_1`, :math:`f_1`.
        matrix2: Parameters :math:`a_2`, :math:`b_2`, :math:`c_2`, :math:`d_2`,
            :math:`e_2`, :math:`f_2`.

    Returns:
        Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
        :math:`f`.
    """
    a1, b1, c1, d1, e1, f1 = matrix1
    a2, b2, c2, d2, e2, f2 = matrix2

    a = a1 * a2 + b1 * d2
    b = a1 * b2 + b1 * e2
    c = a1 * c2 + b1 * f2 + c1
    d = d1 * a2 + e1 * d2
    e = d1 * b2 + e1 * e2
    f = d1 * c2 + e1 * f2 + f1

    return (a, b, c, d, e, f)


def left_matmuls(*matrices: Matrix) -> Matrix:
    r"""Matrix product of :math:`N` matrices from the left

    .. math::

        \begin{pmatrix}
            a_N & b_N & c_N \\
            d_N & e_N & f_N \\
            0   & 0   & 1 \\
        \end{pmatrix}
        \cdot
        \quad\dots\quad
        \cdot
        \begin{pmatrix}
            a_n & b_n & c_n \\
            d_n & e_n & f_n \\
            0   & 0   & 1 \\
        \end{pmatrix}
        \quad\dots\quad
        \cdot
        \begin{pmatrix}
            a_1 & b_1 & c_1 \\
            d_1 & e_1 & f_1 \\
            0   & 0   & 1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}

    Args:
        *matrices: Parameters :math:`a_n`, :math:`b_n`, :math:`c_n`, :math:`d_n`,
            :math:`e_n`, :math:`f_n` of each matrix.

    Returns:
        Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
        :math:`f`.
    """
    return reduce(lambda matrix1, matrix2: matmul(matrix2, matrix1), matrices)


def matinv(matrix: Matrix) -> Matrix:
    r"""Matrix inverse

    .. math::

        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}^{-1}
        =
        \begin{pmatrix}
            a^\prime & b^\prime & c^\prime \\
            d^\prime & e^\prime & f^\prime \\
            0        & 0        & 1 \\
        \end{pmatrix}

    Args:
        matrix: Parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`, :math:`e`,
            :math:`f`.

    Returns:
        Parameters :math:`a^\prime`, :math:`b^\prime`, :math:`c^\prime`,
        :math:`d^\prime`, :math:`e^\prime`, :math:`f^\prime`.
    """
    a, b, c, d, e, f = matrix

    det = a * e - b * d
    ainv = e / det
    binv = -b / det
    cinv = (b * f - c * e) / det
    dinv = -d / det
    einv = a / det
    finv = (c * d - a * f) / det

    return (ainv, binv, cinv, dinv, einv, finv)


def deg2rad(angle_in_deg: float) -> float:
    """Converts an angle from degrees to radians

    Args:
        angle_in_deg: Angle in degrees.

    Returns:
        Angle in radians.
    """
    factor = pi / 180.0
    # FIXME: remove module operation
    return (angle_in_deg % 360.0) * factor


def transform_coordinate(coordinate: Coordinate, matrix: Matrix) -> Coordinate:
    r"""Transforms a ``coordinate`` based on an affine ``matrix``

    .. math::

        \begin{pmatrix}
            a & b & c \\
            d & e & f \\
            0 & 0 & 1 \\
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            x \\
            y \\
            1 \\
        \end{pmatrix}
        =
        \begin{pmatrix}
            x^\prime \\
            y^\prime \\
            1 \\
        \end{pmatrix}

    Args:
        coordinate: Coordinate (:math:`x`, :math:`y`).
        matrix: Affine parameters :math:`a`, :math:`b`, :math:`c`, :math:`d`,
            :math:`e`, :math:`f`.

    Returns:
        Transformed coordinate (:math:`x^\prime`, :math:`y^\prime`).
    """
    x, y = coordinate
    a, b, c, d, e, f = matrix

    xtrans = a * x + b * y + c
    ytrans = d * x + e * y + f

    return (xtrans, ytrans)



import unittest
import random
import numpy as np
from pillow_affine import utils


def randn(mu=0.0, sigma=1.0):
    return random.gauss(mu, sigma)


def random_matrix(seed=None):
    if seed is not None:
        random.seed(seed)
    return tuple([randn() for _ in range(6)])


def convert_matrix_to_numpy(pil_matrix):
    a, b, c, d, e, f = pil_matrix
    return np.array(((a, b, c), (d, e, f), (0.0, 0.0, 1.0)))


class Tester(unittest.TestCase):
    def assertMatrixAlmostEqual(self, matrix1, matrix2, **kwargs):
        def cast(matrix):
            if isinstance(matrix, np.ndarray):
                return matrix

            return convert_matrix_to_numpy(matrix)

        np.testing.assert_allclose(cast(matrix1), cast(matrix2), **kwargs)

    def test_matmul(self):
        random.seed(0)

        pil_matrix1 = random_matrix()
        pil_matrix2 = random_matrix()

        numpy_matrix1 = convert_matrix_to_numpy(pil_matrix1)
        numpy_matrix2 = convert_matrix_to_numpy(pil_matrix2)

        actual = utils.matmul(pil_matrix1, pil_matrix2)
        desired = np.matmul(numpy_matrix1, numpy_matrix2)
        self.assertMatrixAlmostEqual(actual, desired)

    def test_left_matmuls(self):
        random.seed(0)

        pil_matrix1 = random_matrix()
        pil_matrix2 = random_matrix()
        pil_matrix3 = random_matrix()

        numpy_matrix1 = convert_matrix_to_numpy(pil_matrix1)
        numpy_matrix2 = convert_matrix_to_numpy(pil_matrix2)
        numpy_matrix3 = convert_matrix_to_numpy(pil_matrix3)

        actual = utils.left_matmuls(pil_matrix1, pil_matrix2, pil_matrix3)
        desired = np.matmul(numpy_matrix3, np.matmul(numpy_matrix2, numpy_matrix1))
        self.assertMatrixAlmostEqual(actual, desired)

    def test_matinv(self):
        pil_matrix = random_matrix(seed=0)
        numpy_matrix = convert_matrix_to_numpy(pil_matrix)

        actual = utils.matinv(pil_matrix)
        desired = np.linalg.inv(numpy_matrix)
        self.assertMatrixAlmostEqual(actual, desired)

    def test_deg2rad(self):
        angles = (0.0, 90.0, -90.0, 30.0, 360.0, -360.0, 720.0)

        actual = np.array([utils.deg2rad(angle) for angle in angles])
        desired = np.deg2rad(np.array(angles)) % (2 * np.pi)
        np.testing.assert_allclose(actual, desired)

    def test_transform_coordinate(self):
        random.seed(0)

        pil_coordinate = (randn(), randn())
        pil_matrix = random_matrix()

        numpy_coordinate = np.array((*pil_coordinate, 1.0))
        numpy_matrix = convert_matrix_to_numpy(pil_matrix)

        actual = np.array(utils.transform_coordinate(pil_coordinate, pil_matrix))
        desired = np.matmul(numpy_matrix, numpy_coordinate)[:-1]
        np.testing.assert_allclose(actual, desired)



__name__ = "pillow_affine"
__description__ = "Affine transformation utilities for pillow "
__version__ = "0.1"
__url__ = "https://github.com/pmeier/pillow_affine"
__license__ = "BSD-3"
__author__ = "Philip Meier"
__author_email__ = "github.pmeier@posteo.de"



import os
from os import path
from setuptools import find_packages
from importlib import import_module
import itertools
import unittest
import pillow_affine


class Tester(unittest.TestCase):
    def test_import(self):
        pkg_name = "pillow_affine"
        here = path.abspath(path.dirname(__file__))
        pkg_root = path.join(here, "..", pkg_name)

        packages = find_packages(pkg_root)

        modules = []
        for package in packages:
            files = os.listdir(path.join(pkg_root, package.replace(".", os.sep)))
            for file in files:
                name, ext = path.splitext(file)
                if ext == ".py" and name != "__init__":
                    modules.append(f"{package}.{name}")

        for module in itertools.chain(packages, modules):
            import_module(f"{pkg_name}.{module}")

    def test_about(self):
        for attr in (
            "name",
            "description",
            "version",
            "url",
            "license",
            "author",
            "author_email",
        ):
            self.assertIsInstance(getattr(pillow_affine, f"__{attr}__"), str)



from typing import Union, Optional, Sequence, Tuple
from abc import ABC, abstractmethod
from math import floor, ceil
from PIL import Image
from .matrix import (
    shearing_matrix,
    rotation_matrix,
    scaling_matrix,
    translation_matrix,
)
from .utils import Coordinate, Matrix, left_matmuls, matinv, transform_coordinate

__all__ = [
    "AffineTransform",
    "Shear",
    "Rotate",
    "Scale",
    "Translate",
    "ComposedTransform",
]

Size = Tuple[int, int]


def calculate_image_center(size: Size) -> Coordinate:
    """Calculates the center of an image

    Args:
        size: Image size (width, height).

    Returns:
        Image center.
    """
    width, height = size
    horz_center = width / 2.0
    vert_center = height / 2.0
    return horz_center, vert_center


class AffineTransform(ABC):
    """ABC for all affine transformations.
    """

    @abstractmethod
    def _create_matrix(self, size: Size) -> Matrix:
        pass

    def extract_transform_params(
        self, size: Size, expand: bool = False
    ) -> Tuple[Size, int, Matrix]:
        """Extracts the transformation parameters that need to be passed to
        `Image.transform() <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.transform>`_
        for the affine transformation. An simple call might look like::

            from PIL import Image
            from pillow_affine import transforms


            image = Image.open(...)
            transform = transforms.Rotate(30.0)

            transform_params = transform.extract_transform_params(image.size)
            transformed_image = image.transform(*transform_params)

        Args:
            size: Image size (width, height).
            expand: If ``True``, expands the canvas to hold the complete
                transformed motif. Defaults to ``False``.

        .. note::
            If you use the ``expand`` flag the motif is centered on the canvas
            and thus any final translation is removed.

        Returns:
            ``size``, ``method``, and ``data`` parameters for
            `Image.transform() <https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.transform>`_
             .
        """
        transform_matrix = self._create_matrix(size)

        if expand:
            expanded_size, transform_matrix = self._expand_canvas(
                size, transform_matrix
            )
        else:
            expanded_size = size

        transform_matrix = self._coordinate_system_transform(size, transform_matrix)

        data = self._extract_affine_data(transform_matrix)

        return expanded_size, Image.AFFINE, data

    @staticmethod
    def _expand_canvas(size: Size, transform_matrix: Matrix) -> Tuple[Size, Matrix]:
        def calculate_motif_vertices(transform_matrix: Matrix) -> Sequence[Coordinate]:
            width, height = size
            image_vertices = ((0.0, 0.0), (width, 0.0), (0.0, height), (width, height))
            return [
                transform_coordinate(coordinate, transform_matrix)
                for coordinate in image_vertices
            ]

        def calculate_expanded_size(motif_vertices: Sequence[Coordinate]) -> Size:
            xs, ys = zip(*motif_vertices)
            left = floor(min(xs))
            bottom = floor(min(ys))
            right = ceil(max(xs))
            top = ceil(max(ys))

            expanded_width = right - left
            expanded_height = top - bottom
            return expanded_width, expanded_height

        def recenter_motif(expanded_size: Size, transform_matrix: Matrix) -> Matrix:
            matrix = left_matmuls(
                translation_matrix(calculate_image_center(size), inverse=True),
                translation_matrix(calculate_image_center(expanded_size)),
            )
            # TODO: Investigate and document why this extra step is needed
            matrix = AffineTransform._coordinate_system_transform(size, matrix)
            return left_matmuls(transform_matrix, matrix)

        motif_vertices = calculate_motif_vertices(transform_matrix)
        expanded_size = calculate_expanded_size(motif_vertices)
        transform_matrix = recenter_motif(expanded_size, transform_matrix)

        return expanded_size, transform_matrix

    @staticmethod
    def _coordinate_system_transform(size: Size, transform_matrix: Matrix) -> Matrix:
        width, height = size
        matrix = (1.0, 0.0, 0.0, 0.0, -1.0, height)
        return left_matmuls(matrix, transform_matrix, matinv(matrix))

    @staticmethod
    def _extract_affine_data(transform_matrix: Matrix) -> Matrix:
        return matinv(transform_matrix)

    @staticmethod
    def _off_center_transform(
        coordinate: Coordinate, transform_matrix: Matrix
    ) -> Matrix:
        return left_matmuls(
            translation_matrix(coordinate, inverse=True),
            transform_matrix,
            translation_matrix(coordinate, inverse=False),
        )


class ElementaryTransform(AffineTransform):
    @abstractmethod
    def _create_matrix(self, size: Size) -> Matrix:
        pass

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._extra_repr()})"

    def _extra_repr(self) -> str:
        return ""


class Shear(ElementaryTransform):
    """Affine horizontal shearing transformation.

    Args:
        angle: Shearing angle in degrees.
        clockwise: If ``True``, the shearing will be performed clockwise.
            Defaults to ``False``.
        center: Optional center of the shearing. Defaults to the center of
            the image.
    """

    def __init__(
        self,
        angle: float,
        clockwise: bool = False,
        center: Optional[Coordinate] = None,
    ):
        self.angle = angle % 360.0
        self.clockwise = clockwise
        self.center = center

    def _create_matrix(self, size: Size) -> Matrix:
        matrix = shearing_matrix(self.angle, clockwise=self.clockwise)
        if self.center is None:
            center = calculate_image_center(size)
        else:
            center = self.center
        matrix = self._off_center_transform(center, matrix)
        return matrix

    def _extra_repr(self) -> str:
        extras = [f"{self.angle:4.1f}°"]
        if self.clockwise:
            extras.append(f"clockwise={self.clockwise}")
        if self.center is not None:
            extras.append(f"center={self.center}")
        return ", ".join(extras)


class Rotate(ElementaryTransform):
    """Affine rotation transformation.

    Args:
        angle: Rotation angle in degrees.
        clockwise: If ``True``, the rotation will be performed clockwise.
            Defaults to ``False``.
        center: Optional center of the rotation. Defaults to the center of the
            image.
    """

    def __init__(
        self,
        angle: float,
        clockwise: bool = False,
        center: Optional[Coordinate] = None,
    ):
        self.angle = angle % 360.0
        self.clockwise = clockwise
        self.center = center

    def _create_matrix(self, size: Size) -> Matrix:
        matrix = rotation_matrix(self.angle, clockwise=self.clockwise)
        if self.center is None:
            center = calculate_image_center(size)
        else:
            center = self.center
        matrix = self._off_center_transform(center, matrix)
        return matrix

    def _extra_repr(self) -> str:
        extras = [f"{self.angle:4.1f}°"]
        if self.clockwise:
            extras.append(f"clockwise={self.clockwise}")
        if self.center is not None:
            extras.append(f"center={self.center}")
        return ", ".join(extras)


class Scale(ElementaryTransform):
    """Affine scaling transformation

    Args:
        factor: Horizontal and vertical scaling factors. If scalar, the
            same factor is used for both directions.
        center: Optional center of the scaling. Defaults to the center of
            the image.
    """

    def __init__(
        self,
        factor: Union[float, Tuple[float, float]],
        center: Optional[Coordinate] = None,
    ):
        self.factor = factor
        self.center = center

    def _create_matrix(self, size: Size) -> Matrix:
        matrix = scaling_matrix(self.factor)
        if self.center is None:
            center = calculate_image_center(size)
        else:
            center = self.center
        matrix = self._off_center_transform(center, matrix)
        return matrix

    def _extra_repr(self) -> str:
        def format_factor(factor: float) -> str:
            return f"{factor:.2f}"

        extras = []
        if isinstance(self.factor, float):
            extras.append(format_factor(self.factor))
        else:
            horz_factor, vert_factor = [
                format_factor(dim_factor) for dim_factor in self.factor
            ]
            extras.append(f"({horz_factor}, {vert_factor})")
        if self.center is not None:
            extras.append(f"center={self.center}")
        return ", ".join(extras)


class Translate(ElementaryTransform):
    """Affine translation transformation

    Args:
        translation: Horizontal and vertical translation.
        inverse: If ``True``, the translation will be performed in the
            opposite direction. Defaults to ``False``.
    """

    def __init__(self, translation: Coordinate, inverse: bool = False) -> None:
        self.translation = translation
        self.inverse = inverse

    def _create_matrix(self, size: Size) -> Matrix:
        return translation_matrix(self.translation, inverse=self.inverse)

    def _extra_repr(self) -> str:
        extras = [f"{tuple([round(coord, 1) for coord in self.translation])}"]
        if self.inverse:
            extras.append(f"inverse={self.inverse}")
        return ", ".join(extras)


class ComposedTransform(AffineTransform):
    """Composed affine transformation by chaining multiple
    :class:`AffineTransform` s together. An simple example might look like::

        from PIL import Image
        from pillow_affine import transforms

        image = Image.open(...)
        transform = transforms.ComposedTransform(
            transforms.Rotate(30.0), transforms.Translate((50.0, 100.0))
        )

        transform_params = transform.extract_transform_params(image.size)
        transformed_image = image.transform(*transform_params)

    Args:
        transforms: Individual :class:`AffineTransform` s.
    """

    def __init__(self, *transforms: AffineTransform) -> None:
        if len(transforms) == 0:
            msg = "A ComposedTransform must comprise at least one other transform."
            raise RuntimeError(msg)
        self.transforms = transforms

    def _create_matrix(self, size: Size) -> Matrix:
        return left_matmuls(
            *[transform._create_matrix(size) for transform in self.transforms]
        )

    def __repr__(self) -> str:
        head = f"{self.__class__.__name__}("
        tail = ")"

        if len(self.transforms) == 1:
            return head + repr(self.transforms[0]) + tail

        body = [" " * 2 + repr(transform) for transform in self.transforms]
        return "\n".join((head, *body, tail))



from typing import Tuple
from os import path
import unittest
import re
from pyimagetest import ImageTestCase
from pillow_affine import transforms


def convert_angle(angle: float, clockwise: bool = False) -> float:
    return -angle if clockwise else angle


def convert_center(
    center: Tuple[float, float], size: Tuple[int, int]
) -> Tuple[float, float]:
    width, height = size
    return center[0], height - center[1]


def convert_translation(
    translation: Tuple[float, float], inverse: bool = False
) -> Tuple[float, float]:
    translate = translation[0], -translation[1]
    if inverse:
        translate = (-translate[0], -translate[1])
    return translate


class Tester(ImageTestCase):
    def default_image_file(self) -> str:
        here = path.abspath(path.dirname(__file__))
        return path.join(here, "..", "docs", "source", "_static", "images", "raw.png")

    def default_image_backend(self):
        return "PIL"

    def assertIsIdentityTransform(self, transform):
        _, _, actuals = transform.extract_transform_params((1, 1))
        desireds = (1.0, 0.0, 0.0, 0.0, 1.0, 0.0)
        self.assertCountEqual(actuals, desireds)
        for actual, desired in zip(actuals, desireds):
            self.assertAlmostEqual(actual, desired)

    def assertHasValidElementaryTransformRepr(self, transform, special_chars=""):
        name = transform.__class__.__name__
        valid_extra_repr_chars = r"\w\s.,=()" + special_chars
        valid_pattern = name + f"[(][{valid_extra_repr_chars}]*[)]"
        self.assertIsNotNone(re.fullmatch(valid_pattern, repr(transform)))

    def test_AffineTransform(self):
        with self.assertRaises(TypeError):
            transforms.AffineTransform()

    def test_ElementaryTransform(self):
        with self.assertRaises(TypeError):
            transforms.ElementaryTransform()

    def test_Shear_identity(self):
        angle = 0.0

        transform = transforms.Shear(angle)
        self.assertIsIdentityTransform(transform)

    def test_Shear_clockwise_identity(self):
        angle = 0.0
        clockwise = True

        transform = transforms.Shear(angle, clockwise=clockwise)
        self.assertIsIdentityTransform(transform)

    def test_Shear(self):
        angle = 30.0

        transform = transforms.Shear(angle)
        self.assertHasValidElementaryTransformRepr(transform, special_chars="°")

    def test_Shear_clockwise(self):
        angle = 30.0
        clockwise = True

        transform = transforms.Shear(angle, clockwise=clockwise)
        self.assertHasValidElementaryTransformRepr(transform, special_chars="°")

    def test_Shear_off_image_center(self):
        angle = 30.0
        center = (0.0, 0.0)

        transform = transforms.Shear(angle, center=center)
        self.assertHasValidElementaryTransformRepr(transform, special_chars="°")

    def test_Rotate_identity(self):
        angle = 0.0

        transform = transforms.Rotate(angle)
        self.assertIsIdentityTransform(transform)

    def test_Rotate_clockwise_identity(self):
        angle = 0.0
        clockwise = True

        transform = transforms.Rotate(angle, clockwise=clockwise)
        self.assertIsIdentityTransform(transform)

    def test_Rotate(self):
        image = self.load_image()
        angle = 30.0

        transform = transforms.Rotate(angle)
        transform_params = transform.extract_transform_params(image.size)
        actual = image.transform(*transform_params)

        desired = image.rotate(convert_angle(angle))

        self.assertImagesAlmostEqual(actual, desired)
        self.assertHasValidElementaryTransformRepr(transform, special_chars="°")

    def test_Rotate_clockwise(self):
        image = self.load_image()
        angle = 30.0
        clockwise = True

        transform = transforms.Rotate(angle, clockwise=clockwise)
        transform_params = transform.extract_transform_params(image.size)
        actual = image.transform(*transform_params)

        desired = image.rotate(convert_angle(angle, clockwise=clockwise))

        self.assertImagesAlmostEqual(actual, desired)
        self.assertHasValidElementaryTransformRepr(transform, special_chars="°")

    def test_Rotate_off_image_center(self):
        image = self.load_image()
        angle = 30.0
        image_center = transforms.calculate_image_center(image.size)
        center = (image_center[0] - 50.0, image_center[1] + 100.0)

        transform = transforms.Rotate(angle, center=center)
        transform_params = transform.extract_transform_params(image.size)
        actual = image.transform(*transform_params)

        desired = image.rotate(angle, center=convert_center(center, image.size))

        self.assertImagesAlmostEqual(actual, desired)
        self.assertHasValidElementaryTransformRepr(transform, special_chars="°")

    def test_Scale_identity(self):
        factor = 1.0

        transform = transforms.Scale(factor)
        self.assertIsIdentityTransform(transform)

    def test_Scale_async_identity(self):
        factor = (1.0, 1.0)

        transform = transforms.Scale(factor)
        self.assertIsIdentityTransform(transform)

    def test_Scale(self):
        factor = 0.5

        transform = transforms.Scale(factor)
        self.assertHasValidElementaryTransformRepr(transform)

    def test_Scale_async(self):
        factor = (0.5, 2.0)

        transform = transforms.Scale(factor)
        self.assertHasValidElementaryTransformRepr(transform)

    def test_Scale_off_image_center(self):
        factor = 0.5
        center = (0.0, 0.0)

        transform = transforms.Scale(factor, center=center)
        self.assertHasValidElementaryTransformRepr(transform)

    def test_Translate_identity(self):
        translation = (0.0, 0.0)

        transform = transforms.Translate(translation)
        self.assertIsIdentityTransform(transform)

    def test_Translate_inverse_identity(self):
        translation = (0.0, 0.0)
        inverse = True

        transform = transforms.Translate(translation, inverse=inverse)
        self.assertIsIdentityTransform(transform)

    def test_Translate(self):
        image = self.load_image()
        translation = (100.0, 50.0)

        transform = transforms.Translate(translation)
        transform_params = transform.extract_transform_params(image.size)
        actual = image.transform(*transform_params)

        desired = image.rotate(0.0, translate=convert_translation(translation))

        self.assertImagesAlmostEqual(actual, desired)
        self.assertHasValidElementaryTransformRepr(transform)

    def test_Translate_inverse(self):
        image = self.load_image()
        translation = (100.0, 50.0)
        inverse = True

        transform = transforms.Translate(translation, inverse=inverse)
        transform_params = transform.extract_transform_params(image.size)
        actual = image.transform(*transform_params)

        desired = image.rotate(
            0.0, translate=convert_translation(translation, inverse=inverse)
        )

        self.assertImagesAlmostEqual(actual, desired)
        self.assertHasValidElementaryTransformRepr(transform)

    def test_empty_ComposedTransform(self):
        with self.assertRaises(RuntimeError):
            transforms.ComposedTransform()

    def test_ComposedTransform(self):
        image = self.load_image()
        angle = 30.0
        translation = (100.0, 50.0)

        transform = transforms.ComposedTransform(
            transforms.Rotate(angle), transforms.Translate(translation)
        )
        transform_params = transform.extract_transform_params(image.size)
        actual = image.transform(*transform_params)

        angle = convert_angle(angle)
        translate = convert_translation(translation)
        desired = image.rotate(angle, translate=translate)

        self.assertImagesAlmostEqual(actual, desired)

    def test_expand(self):
        image = self.load_image()
        angle = 30.0
        expand = True

        transform = transforms.Rotate(angle)
        transform_params = transform.extract_transform_params(image.size, expand=expand)
        actual = image.transform(*transform_params)

        desired = image.rotate(convert_angle(angle), expand=expand)

        self.assertImagesAlmostEqual(actual, desired)


if __name__ == "__main__":
    unittest.main()
